# {Patman} (Patch Manipulator) is a library for text file patching. It
# can also be used to extract information from files.
class Patman

    # {Patman} exception base class.
    class PatmanError < RuntimeError; end

    # {Patman} searching error (for find functions).
    class PatmanSearchError < PatmanError; end

    # {Patman} file error.
    class PatmanFileError < PatmanError; end


    attr_accessor :marks


    # Create editing session with file.
    def Patman.read( file )
        p = Patman.new( file )
        p.read
        p
    end


    # Edit file and also create it if it does not exist.
    #
    # If block is given, a file will be opened and block with Patman
    # commands will be executed for it. Otherwise a Patman object is
    # returned for further use.
    def Patman.edit( file, &blk )
        p = Patman.new( file )
        if File.exist?( file )
            p.send( :read, file )
        end
        if block_given?
            p.instance_exec( &blk )
            p.send( :write )
        else
            p
        end
    end

    # Create {Patman} object.
    def initialize( file )
        @file = file
        @lines = []
        @line = 0
        @mark = nil
        @marks = {}
        @blockline = nil
        @edited = false
    end

    # Read file in.
    def read( file = @file )
        if File.exist?( file )
            @lines = read_clean( file )
        else
            raise PatmanFileError
        end
    end

    # Write {Patman} content to disk.
    def write( file = @file )
        return unless @edited

        file_dir = File.dirname( file )
        unless File.exist?( file_dir )
            require 'fileutils'
            FileUtils.mkdir_p( file_dir )
        end

        fh = File.open( file, 'w' )

        @lines.each do |line|
            if line
                fh.puts line
            else
                fh.puts ""
            end
        end
        fh.close
        @edited = false
        self
    end

    # Copy {Patman} content to file.
    def copy( file )
        write( file )
        self
    end

    # Return or set line.
    def line( arg = nil )
        if arg
            @line = (arg-1)
            self
        else
            @line+1
        end
    end

    # Step forward or backward current position.
    def step( dir = 1 )
        @line = @line + dir
        self
    end

    # Jump to first line.
    def firstline
        @line = 0
        self
    end

    # Jump to last line.
    def lastline
        @line = @lines.length-1
        self
    end

    # Jump to line after block.
    def blockline
        if @blockline
            @line = @blockline
        end
        self
    end

    # Get or set all {Patman} content.
    def lines( arg = nil )
        if arg
            @edited = true
            @lines = arg
        else
            @lines
        end
    end

    # Reference {Patman} content by range.
    def []( range )
        @lines[ range ]
    end

    # Get current line or lines by count.
    def get( count = 1 )
        if count == 1
            @lines[ @line ]
        else
            @lines[ @line .. (@line+count-1) ]
        end
    end

    # Get current line or any line.
    def ref( line = nil )
        if line
            @lines[ line-1 ]
        else
            @lines[ @line ]
        end
    end

    # Set current line.
    def set( text )
        @edited = true
        @lines[ @line ] = text
        self
    end

    # Substitution in current line.
    def sub( from, to )
        @edited = true
        @lines[ @line ] = @lines[ @line ].sub( from, to )
        self
    end

    # Update current line content (i.e. get&set) with the return value
    # of the given block. Hence last stmt should include the new line
    # content.
    #
    # @example
    #   r.update do |c|
    #      c.sub!( /foo/, 'bar' )
    #      c
    #   end
    #
    def update( &blk )
        @edited = true
        set( yield( get ) )
        self
    end

    # Insert line or lines (Array) to current position.
    def insert( text = nil )
        @edited = true
        if text.kind_of? Array
            line = @line
            step( -1 )
            text.each do |txt|
                append( txt )
            end
            @line = line
        else
            @lines.insert( @line, text )
        end
        self
    end

    # Append after current position.
    def append( text = nil )
        @edited = true
        if text.kind_of? Array
            text.each do |txt|
                append( txt )
            end
        else
            step
            @lines.insert( @line, text )
        end
        self
    end

    # Delete current line.
    def delete( count = 1 )
        @edited = true
        count.times do |i|
            @lines.delete_at( @line )
        end
        self
    end

    # Insert file to current position.
    def insertfile( file )
        @edited = true
        step( -1 )
        read_clean( file ).each do |line|
            append line
        end
        self
    end

    # Clear {Patman} content and reset current line.
    def clear
        @edited = true
        @lines = []
        @line = 0
        self
    end

    # Find Regexp or literal string forwards or backwards. Return true
    # on success.
    def find( re_or_str, forward = true )
        begin
            @line = search_with_exception( re_or_str, forward )
            true
        rescue
            false
        end
    end

    # Search Regexp or literal string forwards or backwards. Fail with
    # expection (PatmanSearchError) if not found.
    def search( re_or_str, forward = true )
        @line = search_with_exception( re_or_str, forward )
        self
    end

    # Return line count of {Patman} content.
    def length
        @lines.length
    end

    # Return {Patman} file name.
    def filename
        @file
    end

    # Mark content modified (explicit).
    def edit
        @edited = true
    end

    # Return true if content is modified.
    def edited?
        @edited
    end

    # Execute block, retain current position, and return block value.
    def excursion( &blk )
        line = @line
        ret = instance_eval( &blk )
        @line = line
        ret
    end

    # Mark (store) current position to default or to named mark.
    def mark( tag = nil )
        if tag
            @marks[ tag ] = @line+1
            self
        else
            @mark = @line+1
            self
        end
    end

    # Unmark (restore) current position from default or from named
    # mark.
    def unmark( tag = nil )
        if tag && @marks[ tag ]
            @line = @marks[ tag ]-1
            self
        elsif @mark
            @line = @mark-1
            @mark = nil
            self
        else
            self
        end
    end

    # Execute given block for all lines, i.e. all positions. Block
    # parameter is {Patman}.
    def do_all( &blk )
        do_for( 1, length-1, &blk )
    end

    # Execute given block between start and stop positions, and update
    # position.
    def do_range( start, stop, &blk )
        do_for( start, (stop-start+1), &blk )
    end

    # Execute given block starting from start by count, and update
    # position.
    def do_for( start, count, &blk )
        line = @line
        @line = start-1
        count.times do
            yield self
            @line += 1
        end
        @blockline = @line
        @line = line
        self
    end

    # Get lines between start and stop positions inclusive.
    def get_range( start, stop )
        @lines[ (start-1) .. (stop-1) ]
    end

    # Get lines starting from start by count.
    def get_for( start, count )
        @lines[ (start-1) ... (start-1+count) ]
    end

    # View line content around current position (by count).
    def peek( count = 0 )
        view_range( @line-count, @line+count+1 )
        nil
    end

    # View line content with line numbers around current position (by
    # count).
    def peek_ln( count = 0 )
        view_range( @line-count, @line+count+1, true )
        nil
    end

    # View line content.
    #
    # * no args:  view all
    # * one arg:  view from current onwards by count
    # * two args: view given range
    def view( arg1 = nil, arg2 = nil )
        if !arg1 && !arg2
            view_range( 0, length )
        elsif arg1 && !arg2
            view_range( @line, @line+arg1 )
        elsif arg1 && arg2
            view_range( arg1-1, arg1-1+arg2 )
        end
        nil
    end

    # View line content with line numbers.
    #
    # * no args:  view all
    # * one arg:  view from current onwards by count
    # * two args: view given range
    def view_ln( arg1 = nil, arg2 = nil )
        if !arg1 && !arg2
            view_range( 0, length, true )
        elsif arg1 && !arg2
            view_range( @line, @line+arg1, true )
        elsif arg1 && arg2
            view_range( arg1-1, arg1-1+arg2, true )
        end
        nil
    end



    private

    def read_clean( file ) # :nodoc:
        File.readlines( file ).map{|line| line.chomp}.map{|line| line.empty? ? nil : line}
    end

    def view_range( first, last, ln = false ) # :nodoc:
        if ln
            ln = first+1
            @lines[first...last].each do |line|
                puts "#{ln.to_s.rjust(3)}: #{line}"
                ln += 1
            end
        else
            @lines[first...last].each do |line|
                puts line
            end
        end
    end

    def search_with_exception( re_or_str, forward ) # :nodoc:

        line = @line
        len = @lines.length

        if re_or_str.kind_of? String

            if forward
                while line < len
                    if @lines[line] && @lines[line].include?( re_or_str )
                        return line
                    end
                    line += 1
                end
            else
                while line >= 0
                    if @lines[line] && @lines[line].include?( re_or_str )
                        return line
                    end
                    line -= 1
                end
            end

        else

            if forward
                while line < len
                    if @lines[line] && re_or_str.match( @lines[line] )
                        return line
                    end
                    line += 1
                end
            else
                while line >= 0
                    if @lines[line] && re_or_str.match( @lines[line] )
                        return line
                    end
                    line -= 1
                end
            end
        end

        raise PatmanSearchError, "#{@file}: Could not find \"#{re_or_str}\"..."
    end

end

require_relative 'version'
